# Copyright 2021 Juan L Gamella

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""Module containing the DecomposableScore class, inherited by all
classes which implement a locally decomposable score for directed
acyclic graphs. By default, the class also caches the results of
computing local scores.
NOTE: It is not mandatory to inherit this class when developing custom
scores to use with the GES implementation in ges.py. The only
requirement is that the class defines:
  1. the local_score function (see below),
  2. an attribute "p" for the total number of variables.
"""

import copy

# --------------------------------------------------------------------
# l0-penalized Gaussian log-likelihood score for a sample from a single
# (observational) environment


class DecomposableScore:
    def __init__(self, data, cache=True, debug=0):
        self._data = copy.deepcopy(data)
        self._cache = {} if cache else None
        self._debug = debug
        self.p = None

    def local(self, x, pa):
        """
        Return the local score of a given node and a set of
        parents. If self.cache=True, will use previously computed
        score if possible.
        Parameters
        ----------
        x : int
            a node
        pa : set of ints
            the node's parents
        Returns
        -------
        score : float
            the corresponding score
        """
        if self._cache is None:
            return self._compute_local_score(x, pa)
        else:
            key = (x, tuple(sorted(pa)))
            try:
                score = self._cache[key]
                print(
                    "score%s: using cached value %0.2f" % (key, score)
                ) if self._debug >= 2 else None
            except KeyError:
                score = self._compute_local_score(x, pa)
                self._cache[key] = score
                print("score%s = %0.2f" % (key, score)) if self._debug >= 2 else None
            return score

    def _compute_local_score(self, x, pa):
        """
        Compute the local score of a given node and a set of
        parents.
        Parameters
        ----------
        x : int
            a node
        pa : set of ints
            the node's parents
        Returns
        -------
        score : float
            the corresponding score
        """
        return 0



"""Various types and representations of graphs."""
import numpy as np


class UndirectedDependenceGraph(object):
    r"""Adjacency matrix representation using a 2d numpy array.

    Upon initialization, this class is fairly standard implementation
    of an undirected graph. However, upon calling the :meth:`make_aux`
    method, an auxilliary data structure in the form of several new
    attributes is created, which are used by
    :meth:`medil.ecc_algorithms.find_clique_min_cover` according to
    the algorithm in :cite:`Gramm_2009`.

    Attributes
    ----------


    Notes
    -----
    The algorithms for finding the minMCM via ECC contain many
    algebraic operations, so adjacency matrix representation (via
    NumPy) is most covenient.

    """

    def __init__(self, adj_matrix, verbose=False):
        # doesn't behave well unless input is nparray
        self.adj_matrix = adj_matrix
        self.num_vertices = np.trace(adj_matrix)
        self.max_num_verts = len(adj_matrix)
        self.num_edges = np.triu(adj_matrix, 1).sum()
        self.verbose = verbose

    def add_edges(self, edges):
        v_1s = edges[:, 0]
        v_2s = edges[:, 1]
        self.adj_matrix[v_1s, v_2s] = 1
        self.adj_matrix[v_2s, v_1s] = 1
        self.num_edges = np.triu(self.adj_matrix, 1).sum()

    def rm_edges(self, edges):
        v_1s = edges[:, 0]
        v_2s = edges[:, 1]
        self.adj_matrix[v_1s, v_2s] = 0
        self.adj_matrix[v_2s, v_1s] = 0
        self.num_edges = np.triu(self.adj_matrix, 1).sum()

    def make_aux(self):
        # this makes the auxilliary structure described in INITIALIZATION in the paper

        # find neighbourhood for each vertex
        # each row corresponds to a unique edge
        max_num_edges = self.n_choose_2(self.max_num_verts)
        self.common_neighbors = np.zeros(
            (max_num_edges, self.max_num_verts), int
        )  # init

        # mapping of edges to unique row idx
        triu_idx = np.triu_indices(self.max_num_verts, 1)
        nghbrhd_idx = np.zeros((self.max_num_verts, self.max_num_verts), int)
        nghbrhd_idx[triu_idx] = np.arange(max_num_edges)
        # nghbrhd_idx += nghbrhd_idx.T
        self.get_idx = lambda edge: nghbrhd_idx[edge[0], edge[1]]

        # reverse mapping
        u, v = np.where(np.triu(np.ones_like(self.adj_matrix), 1))
        self.get_edge = lambda idx: (u[idx], v[idx])

        # compute actual neighborhood for each edge = (v_1, v_2)
        self.nbrs = lambda edge: np.logical_and(
            self.adj_matrix[edge[0]], self.adj_matrix[edge[1]]
        )

        extant_edges = np.transpose(np.triu(self.adj_matrix, 1).nonzero())
        self.extant_edges_idx = np.fromiter(
            {self.get_idx(edge) for edge in extant_edges}, dtype=int
        )
        extant_nbrs = np.array([self.nbrs(edge) for edge in extant_edges], int)
        extant_nbrs_idx = np.array([self.get_idx(edge) for edge in extant_edges], int)

        # from paper: set of N_{u, v} for all edges (u, v)
        self.common_neighbors[extant_nbrs_idx] = extant_nbrs

        # number of cliques for each node? assignments? if we set diag=0
        # num_cliques = common_neighbors.sum(0)

        # sum() of submatrix of graph containing exactly the rows/columns
        # corresponding to the nodes in common_neighbors(edge) using
        # logical indexing:

        # make mask to identify subgraph (closed common neighborhood of
        # nodes u, v in edge u,v)
        mask = lambda edge_idx: np.array(self.common_neighbors[edge_idx], dtype=bool)

        # make subgraph-adjacency matrix, and then subtract diag and
        # divide by two to get num edges in subgraph---same as sum() of
        # triu(subgraph-adjacency matrix) but probably a bit faster
        nbrhood = lambda edge_idx: self.adj_matrix[mask(edge_idx)][:, mask(edge_idx)]
        max_num_edges_in_nbrhood = (
            lambda edge_idx: (nbrhood(edge_idx).sum() - mask(edge_idx).sum()) // 2
        )

        # from paper: set of c_{u, v} for all edges (u, v)
        self.nbrhood_edge_counts = np.array(
            [
                max_num_edges_in_nbrhood(edge_idx)
                for edge_idx in np.arange(max_num_edges)
            ],
            int,
        )

        # important structs are:
        # self.common_neighbors
        # self.nbrhood_edge_counts
        # # and fun is
        # self.nbrs

    @staticmethod
    def n_choose_2(n):
        return n * (n - 1) // 2

    def reducible_copy(self):
        return ReducibleUndDepGraph(self)


class ReducibleUndDepGraph(UndirectedDependenceGraph):
    def __init__(self, udg):
        self.unreduced = udg.unreduced if hasattr(udg, "unreduced") else udg
        self.adj_matrix = udg.adj_matrix.copy()
        self.num_vertices = udg.num_vertices
        self.num_edges = udg.num_edges

        self.the_cover = None
        self.verbose = udg.verbose

        # from auxilliary structure if needed
        if not hasattr(udg, "get_idx"):
            udg.make_aux()
        self.get_idx = udg.get_idx

        # need to also update these all when self.cover_edges() is called? already done in rule_1
        self.common_neighbors = udg.common_neighbors.copy()
        self.nbrhood_edge_counts = udg.nbrhood_edge_counts.copy()

        # update when cover_edges() is called, actually maybe just extant_edges?
        self.extant_edges_idx = udg.extant_edges_idx.copy()
        self.nbrs = udg.nbrs
        self.get_edge = udg.get_edge

        if hasattr(udg, "reduced_away"):  # then rule_3 wasn't applied
            self.reduced_away = udg.reduced_away

    def reset(self):
        self.__init__(self.unreduced)

    def reduzieren(self, k_num_cliques):
        # reduce by first applying rule 1 and then repeatedly applying
        # rule 2 or rule 3 (both of which followed by rule 1 again)
        # until they don't apply
        if self.verbose:
            print("\t\treducing:")
        self.k_num_cliques = k_num_cliques
        self.reducing = True
        while self.reducing:
            self.reducing = False
            self.rule_1()
            self.rule_2()
            if self.k_num_cliques < 0:
                return
            if self.reducing:
                continue
            self.rule_3()

    def rule_1(self):
        # rule_1: Remove isolated vertices and vertices that are only
        # adjacent to covered edges

        isolated_verts = np.where(self.adj_matrix.sum(0) + self.adj_matrix.sum(1) == 2)[
            0
        ]
        if len(isolated_verts) > 0:  # then Rule 1 is applied
            if self.verbose:
                print("\t\t\tapplying Rule 1...")

            # update auxilliary attributes; LEMMA 2
            self.adj_matrix[isolated_verts, isolated_verts] = 0
            self.num_vertices -= len(isolated_verts)

            # decrease nbrhood edge counts
            for vert in isolated_verts:
                open_nbrhood = self.unreduced.adj_matrix[vert].copy()
                open_nbrhood[vert] = 0
                idx_nbrhoods_to_update = np.where(self.common_neighbors[:, vert] == 1)[
                    0
                ]
                tiled = np.tile(
                    open_nbrhood, (len(idx_nbrhoods_to_update), 1)
                )  # instead of another loop
                to_subtract = np.logical_and(
                    tiled, self.common_neighbors[idx_nbrhoods_to_update]
                ).sum(1)
                self.nbrhood_edge_counts[idx_nbrhoods_to_update] -= to_subtract
                # my own addition:
                # self.nbrhood[:, vert] = 0

            # remove isolated_verts from common neighborhoods
            self.common_neighbors[:, isolated_verts] = 0

        # max_num_edges = self.n_choose_2(self.unreduced.max_num_verts)
        # mask = lambda edge_idx: np.array(self.common_neighbors[edge_idx], dtype=bool)

        # # make subgraph-adjacency matrix, and then subtract diag and
        # # divide by two to get num edges in subgraph---same as sum() of
        # # triu(subgraph-adjacency matrix) but probably a bit faster
        # nbrhood = lambda edge_idx: self.adj_matrix[mask(edge_idx)][:, mask(edge_idx)]
        # max_num_edges_in_nbrhood = lambda edge_idx: (nbrhood(edge_idx).sum() - mask(edge_idx).sum()) // 2

        # # from paper: set of c_{u, v} for all edges (u, v)
        # self.nbrhood_edge_counts = np.array([max_num_edges_in_nbrhood(edge_idx) for edge_idx in np.arange(max_num_edges)], int)
        # # assert (nbrhood_edge_counts==self.nbrhood_edge_counts).all()
        # # print(nbrhood_edge_counts, self.nbrhood_edge_counts)
        # # need to fix!!!!!!!! update isn't working; so just recomputing for now
        # # # # # # # # but actually update produces correct result though recomputing doesn't?

    def rule_2(self):
        # rule_2: If an uncovered edge {u,v} is contained in exactly
        # one maximal clique C, i.e., the common neighbors of u and v
        # induce a clique, then add C to the solution, mark its edges
        # as covered, and decrease k by one

        score = self.n_choose_2(self.common_neighbors.sum(1)) - self.nbrhood_edge_counts
        # score of n implies edge is in exactly n+1 maximal cliques,
        # so we want edges with score 0

        clique_idxs = np.where(score[self.extant_edges_idx] == 0)[0]

        if clique_idxs.size > 0:
            clique_idx = clique_idxs[-1]
            if self.verbose:
                print("\t\t\tapplying Rule 2...")
            clique = self.common_neighbors[self.extant_edges_idx[clique_idx]].copy()
            self.the_cover = (
                clique.reshape(1, -1)
                if self.the_cover is None
                else np.vstack((self.the_cover, clique))
            )
            self.cover_edges()
            self.k_num_cliques -= 1
            self.reducing = True
        # start the loop over so Rule 1 can 'clean up'
        # self.common_neighbors[clique_idxs[0]] = 0  # zero out row, to update struct? not in paper?

    def rule_3(self):
        # rule_3: Consider a vertex v that has at least one
        # prisoner. If each prisoner is connected to at least one
        # vertex other than v via an uncovered edge (automatically
        # given if instance is reduced w.r.t. rules 1 and 2), and the
        # prisoners dominate the exit,s then delete v. To reconstruct
        # a solution for the unreduced instance, add v to every clique
        # containing a prisoner of v.

        for vert, nbrhood in enumerate(self.adj_matrix):
            if nbrhood[vert] == 0:  # then nbrhood is empty
                continue
            exits = np.zeros(self.unreduced.max_num_verts, bool)
            nbrs = np.flatnonzero(nbrhood)
            for nbr in nbrs:
                if (nbrhood.astype(int) - self.adj_matrix[nbr].astype(int) == -1).any():
                    exits[nbr] = True
            # exits[j] == True iff j is an exit for vert
            # prisoners[j] == True iff j is a prisoner of vert
            prisoners = np.logical_and(~exits, self.adj_matrix[vert])
            prisoners[vert] = False  # a vert isn't its own prisoner

            # check if each exit is adjacent to at least one prisoner
            prisoners_dominate_exits = self.adj_matrix[prisoners][:, exits].sum(0)
            if prisoners_dominate_exits.all():  # apply the rule
                self.reducing = True
                if self.verbose:
                    print("\t\t\tapplying Rule 3...")

                # mark edges covered so rule_1 will delete vertex
                edges = np.array(
                    [
                        [vert, u]
                        for u in np.flatnonzero(self.adj_matrix[vert])
                        if vert != u
                    ]
                )
                edges.sort()  # so they're in triu, so that self.get_idx works
                self.cover_edges(edges)

                # keep track of deleted nodes and their prisoners for reconstructing solution
                if not hasattr(self, "reduced_away"):
                    self.reduced_away = np.zeros_like(self.adj_matrix, bool)
                self.reduced_away[vert] = prisoners
                break

    def choose_nbrhood(self):
        # only compute for existing edges
        common_neighbors = self.common_neighbors[self.extant_edges_idx]
        nbrhood_edge_counts = self.nbrhood_edge_counts[self.extant_edges_idx]
        score = self.n_choose_2(common_neighbors.sum(1)) - nbrhood_edge_counts

        # idx in reduced idx list, not from full edge list
        chosen_edge_idx = score.argmin()

        mask = common_neighbors[chosen_edge_idx].astype(bool)

        chosen_nbrhood = self.adj_matrix.copy()
        chosen_nbrhood[~mask, :] = 0
        chosen_nbrhood[:, ~mask] = 0
        if chosen_nbrhood.sum() <= mask.sum():
            raise ValueError("This nbrhood has no edges")
        return chosen_nbrhood

    def cover_edges(self, edges=None):
        if edges is None:
            # always call after updating the cover; only on single, recently added clique
            if self.the_cover is None:
                return  # self.adj_matrix

            # change edges to 0 if they're covered
            clique = self.the_cover[-1]
            covered = np.where(clique)[0]

            # trick for getting combinations from idx
            comb_idx = np.triu_indices(len(covered), 1)

            # actual pairwise combinations; ie all edges (v_i, v_j) covered by the clique
            covered_edges = np.empty((len(comb_idx[0]), 2), int)
            covered_edges[:, 0] = covered[comb_idx[0]]
            covered_edges[:, 1] = covered[comb_idx[1]]

        else:
            covered_edges = edges

        # cover (remove from reduced_graph) edges
        self.rm_edges(covered_edges)
        # update extant_edges_idx
        rmed_edges_idx = [self.get_idx(edge) for edge in covered_edges]
        extant_rmed_edges_idx = [
            edge for edge in rmed_edges_idx if edge in self.extant_edges_idx
        ]
        idx_idx = np.array(
            [np.where(self.extant_edges_idx == idx) for idx in extant_rmed_edges_idx],
            int,
        ).flatten()

        self.extant_edges_idx = np.delete(self.extant_edges_idx, idx_idx)
        # now here do all the updates to nbrs?----actually probably don't want this? see 2clique house example

        # update self.common_neighbors
        # self.common_neighbors[rmed_edges_idx] = 0   # zero out rows covered edges; maybe not necessary, since common_neighbors is (probably?) only called with extant_edges_idx?

        if self.verbose:
            print("\t\t\t{} uncovered edges remaining".format(self.num_edges))

        # if edges is None:
        #     cover_orig = self.the_cover
        #     while self.the_cover.shape[0] > 1:
        #         self.the_cover = self.the_cover[:-1]
        #         self.cover_edges()
        #     self.the_cover = cover_orig

    def reconstruct_cover(self, the_cover):
        if not hasattr(self, "reduced_away"):  # then rule_3 wasn't applied
            return the_cover

        # add the reduced away vert to all covering cliques containing at least one of its prisoners
        to_expand = np.flatnonzero(self.reduced_away.sum(1))
        for vert in to_expand:
            prisoners = self.reduced_away[vert]
            tiled_prisoners = np.tile(
                prisoners, (len(the_cover), 1)
            )  # instead of another loop
            cliques_to_update_mask = (
                np.logical_and(tiled_prisoners, the_cover).sum(1).astype(bool)
            )
            the_cover[cliques_to_update_mask, vert] = 1

        return the_cover


# class minMCM(object):
# TODO: implement as a bigraph with biadjacency matrix with rows M and cols L


# class MCM(object):
# TODO: implement as large DAG adj matrix over L and M, or smaller bigraph for L-M connections and DAG adj matirx for L->L connections



"""Example graphs and data, for use in testing and tutorials."""
import numpy as np


class ExampleUDGAndMCM(object):
    r"""Example consisting of a description, UDG, and MCM"""
    def __init__(self, description):
        self.description = description
        self.UDG = None
        self.MCM = None

    def add_udg(self, udg):
        self.UDG = np.array(udg)

    def add_mcm(self, mcm):
        self.MCM = np.array(mcm)


simple_M = ExampleUDGAndMCM("M-shaped mcm, with 2 latent- and 3 measurement-variables")
simple_M.add_udg(
    [
        [1, 1, 0],
        [1, 1, 1],
        [0, 1, 1]
    ]
)
simple_M.add_mcm(
    [
        [0, 1, 1],
        [1, 1, 0]
    ]
)

triangle = ExampleUDGAndMCM("triangle example, where minECC differs from set of all maximal cliques")
triangle.add_udg(
    [
        [1, 1, 1, 0, 0, 0],
        [1, 1, 1, 1, 1, 0],
        [1, 1, 1, 0, 1, 1],
        [0, 1, 0, 1, 1, 0],
        [0, 1, 1, 1, 1, 1],
        [0, 0, 1, 0, 1, 1],
    ]
)
triangle.add_mcm(
    [
        [0, 0, 1, 0, 1, 1],
        [0, 1, 0, 1, 1, 0],
        [1, 1, 1, 0, 0, 0]
    ]
)

more_latents = ExampleUDGAndMCM("example where number of latent variables is larger than number of measurement variables")
more_latents.add_udg(
    [
        [1, 1, 0, 1, 0, 1],
        [1, 1, 1, 0, 1, 0],
        [0, 1, 1, 1, 0, 0],
        [1, 0, 1, 1, 1, 0],
        [0, 1, 0, 1, 1, 1],
        [1, 0, 0, 0, 1, 1],
    ]
)
more_latents.add_mcm(
    [
        [1, 1, 0, 0, 0, 0],
        [0, 1, 1, 0, 0, 0],
        [0, 0, 1, 1, 0, 0],
        [0, 0, 0, 1, 1, 0],
        [0, 0, 0, 0, 1, 1],
        [1, 0, 0, 0, 0, 1],
        [1, 0, 0, 1, 0, 0],
        [0, 1, 0, 0, 1, 0],
    ]
)

am_cm_diff = ExampleUDGAndMCM("example where multiple ECCs are possible depending on edge minimal vs vertex minimal")
am_cm_diff.add_udg(
    [
        [1, 1, 1, 1, 1, 0, 1, 0],
        [1, 1, 1, 0, 1, 1, 1, 1],
        [1, 1, 1, 0, 1, 1, 1, 1],
        [1, 0, 0, 1, 0, 1, 1, 1],
        [1, 1, 1, 0, 1, 1, 0, 1],
        [0, 1, 1, 1, 1, 1, 0, 1],
        [1, 1, 1, 1, 0, 0, 1, 1],
        [0, 1, 1, 1, 1, 1, 1, 1],
    ]
)
am_cm_diff.add_mcm(
    [
        [1, 1, 1, 0, 1, 0, 0, 0],
        [1, 0, 0, 1, 0, 0, 1, 0],
        [0, 1, 1, 0, 1, 1, 0, 1],
        [0, 0, 0, 1, 0, 1, 0, 1],
        [0, 1, 1, 0, 0, 0, 1, 1],
    ]
)

examples = (simple_M, triangle, more_latents, am_cm_diff)


## Data
# TODO: add linear and nonlinear data example



"""For simulating and learning functional MeDIL Causal Models"""
import torch
import torch.nn as nn
import torch.distributions as dist
from pytorch_lightning import LightningModule
from torch.utils.data import DataLoader
import numpy as np
from collections import OrderedDict


def gaussian_mixture_sampler(
    num_latent, num_mixtures=4, weights=None, means=None, cov=None
):
    """

    :param num_latent:
    :param num_mixtures:
    :param weights:
    :param means:
    :param cov:
    :return:
    """

    if weights is None:
        weights = torch.randn(num_latent, num_mixtures).softmax(dim=1)

    if means is None:
        means = torch.randn(num_latent, num_mixtures, 1) * 2

    if cov is None:
        cov = torch.randn(num_latent, num_mixtures, 1)

    mix = dist.Categorical(weights)
    comp = dist.Independent(dist.Normal(means, cov), 1)

    gmm = dist.MixtureSameFamily(mix, comp)

    return lambda n: gmm.sample((n,)).squeeze()


def uniform_sampler(num_latent, low=-1, high=1):
    """

    :param num_latent:
    :param min:
    :param max:
    :return:
    """

    return lambda n: torch.empty((n, num_latent)).uniform_(low, high)


class NNMechanism(nn.Module):
    def __init__(
        self,
        num_causes,
        num_hidden_layers=1,
        num_hidden_units=20,
        non_linearity="tanh",
        output=None,
        noise_function=None,
        noise_coeff=1.0,
    ):
        """

        :param num_causes:
        :param num_hidden_layers:
        :param num_hidden_units:
        :param activation:
        :param output:
        :param noise_function:
        :param noise_apply:
        """

        super().__init__()
        self.num_causes = num_causes

        # Init Layer list.
        layers = []

        # Input Layer
        layers.append(nn.Linear(self.num_causes + 1, num_hidden_units))

        # Select Activation function
        activation = nn.Identity

        if non_linearity is "tanh":
            activation = nn.Tanh
        elif non_linearity is "relu":
            activation = nn.ReLU
        elif non_linearity is "sigmoid":
            activation = nn.Sigmoid
        elif non_linearity is "gelu":
            activation = nn.GELU

        # Initialise Layers
        for idx in range(num_hidden_layers):
            layers.append(activation())
            layers.append(nn.Linear(num_hidden_units, num_hidden_units))

        # Output Layer
        final = nn.Identity
        if output is "tanh":
            final = nn.Tanh
        elif output is "sigmoid":
            final = nn.Sigmoid

        layers.append(activation())
        layers.append(nn.Linear(num_hidden_units, 1))
        layers.append(final())

        # Assign Layers to Causal Function
        self.causal_function = nn.Sequential(*layers)

        # Horrible shameful hack
        # final_layer_var = 5.0
        # nn.init.uniform_(layers[-2].weight, -final_layer_var, final_layer_var)
        # nn.init.uniform_(layers[-2].bias, -final_layer_var, final_layer_var)
        # print('final', layers[-2].weight, 'pre-final',layers[-4].weight)

        # Exogenous Noise function
        self.noise_function = lambda n: dist.Normal(0, 0.1).sample((n, 1))

        if noise_function is not None:
            self.noise_function = noise_function

        # Exogenous Noise Coefficient
        self.noise_coeff = noise_coeff

    def forward(self, x):
        """

        :param x:
        :return:
        """
        # Sample Exogenous noise corresponding to variable.
        noise = self.noise_coeff * self.noise_function(x.shape[0])

        # Concatenate with input
        causes = torch.cat((noise, x), axis=1)

        # Compute Output
        output = self.causal_function(causes)

        return output


class MeDILCausalModel(nn.Module):
    def __init__(
        self,
        biadj_mat,
        num_hidden_layers=1,
        num_hidden_units=20,
        non_linearity="tanh",
        output="none",
        noise_function=None,
        noise_coeff=1.0,
        debug=False,
    ):

        super().__init__()

        self.biadj_mat = biadj_mat
        self.debug = debug

        self.hparams = {
            "num_hidden_layers": num_hidden_layers,
            "num_hidden_units": num_hidden_units,
            "non_linearity": non_linearity,
            "output": output,
            "noise_coeff": noise_coeff,
        }

        self.num_latent = biadj_mat.shape[0]
        self.num_obs = biadj_mat.shape[1]

        self.subset_mask = [
            np.argwhere(self.biadj_mat[:, idx]).squeeze(axis=-1)
            for idx in range(self.num_obs)
        ]

        print(
            "Generator initialised. #latent: {}, #observed: {}".format(
                self.num_latent, self.num_obs
            )
        )

        forward_dict = {}

        for idx in range(self.num_obs):

            num_latent_causes = np.sum(self.biadj_mat[:, idx].astype(int))

            mechanism = NNMechanism(
                num_latent_causes,
                num_hidden_layers,
                num_hidden_units,
                non_linearity,
                output,
                noise_function,
                noise_coeff,
            )

            forward_dict[str(idx)] = mechanism

        if self.debug:
            print("blah")
        self.observed = nn.ModuleDict(forward_dict)

        if self.debug:
            print("blah")

    def forward(self, x):

        output = []

        if self.debug:
            print("input", x.shape)
        for idx in range(self.num_obs):

            if self.debug:
                print("obs" + str(idx))
            input_subset_mask = np.argwhere(self.biadj_mat[:, idx])

            if self.debug:
                print(input_subset_mask.squeeze())
                print("pre", self.subset_mask[idx])
            input_subset = x[:, self.subset_mask[idx]]

            if self.debug:
                print("input_subset", input_subset.shape)
                print(list(self.observed[str(idx)].modules()))
            obs = self.observed[str(idx)](input_subset)

            if self.debug:
                print("obs", obs.shape)
            output.append(obs)

        output = torch.cat(output, axis=1)

        return output

    def sample(self, sampler, num_samples=100):
        """

        :param sampler:
        :param num_samples:
        :return:
        """

        input_sample = sampler(num_samples)

        self.eval()
        with torch.no_grad():
            output_sample = self.forward(input_sample)

        return input_sample, output_sample

    def set_noise_coeff(self, value):
        """

        :param value:
        :return:
        """

        for module in self.observed:
            # print('before',self.observed[module].noise_coeff)
            self.observed[module].noise_coeff = value
            # print('after',self.observed[module].noise_coeff)

    def interpolate(self, num_samples=100):
        """

        :param num_samples:
        :return:
        """

        self.eval()
        self.set_noise_coeff(0.0)

        output_sample_list = list()

        for latent_idx in range(self.num_latent):

            latent_sample = torch.zeros((num_samples, self.num_latent))
            latent_sample[:, latent_idx] = torch.arange(-10.0, 10.0, 20.0 / num_samples)

            output_sample = self.forward(latent_sample)

            output_sample_list.append(output_sample)

        self.set_noise_coeff(1.0)

        return output_sample_list


class GAN(LightningModule):
    def __init__(self, data_filepath, decoder, latent_sampler=None, batch_size=256):
        """

        :param data_filepath:
        :param decoder:
        :param latent_sampler:
        :param batch_size:
        """

        super().__init__()

        self.hparams = decoder.hparams
        self.hparams["batch_size"] = batch_size
        print(self.hparams)

        self.data_filepath = data_filepath
        self.decoder = decoder
        self.latent_sampler = latent_sampler
        self.batch_size = batch_size

        self.num_latent = self.decoder.num_latent
        self.num_obs = self.decoder.num_obs

    def forward(self, z):

        return self.decoder(z)

    def mmd_loss(self, x_hat_batch, x_batch, alpha=1):

        return MMDLoss(x_batch.shape[0])(x_hat_batch, x_batch)

    def train_dataloader(self):

        dataset = np.load(self.data_filepath).astype(np.float32)

        return DataLoader(dataset, batch_size=self.batch_size, shuffle=True)

    def configure_optimizers(self):
        return torch.optim.Adam(self.parameters())

    # Train Functions
    def training_step(self, batch, batch_idx):

        sample = batch

        z = self.latent_sampler(batch.shape[0])

        x_hat = self.forward(z)

        loss = self.mmd_loss(x_hat, batch)

        tqdm_dict = {"loss": loss}
        output = OrderedDict(
            {
                "loss": loss,
                # 'progress_bar': tqdm_dict,
                "log": tqdm_dict,
            }
        )

        self.logger.log_metrics({"mmd_loss": loss.item()}, batch_idx)

        return output


class MMDLoss(nn.Module):
    def __init__(self, input_size, bandwidths=None):
        """Init the model."""
        super(MMDLoss, self).__init__()
        if bandwidths is None:
            bandwidths = torch.Tensor([0.01, 0.1, 1, 10, 100])
        else:
            bandwidths = bandwidths
        s = torch.cat(
            [
                torch.ones([input_size, 1]) / input_size,
                torch.ones([input_size, 1]) / -input_size,
            ],
            0,
        )

        self.register_buffer("bandwidths", bandwidths.unsqueeze(0).unsqueeze(0))
        self.register_buffer("S", (s @ s.t()))

    def forward(self, x, y):

        X = torch.cat([x, y], 0)

        XX = X @ X.t()
        X2 = (X * X).sum(dim=1).unsqueeze(0)

        exponent = -2 * XX + X2.expand_as(XX) + X2.t().expand_as(XX)

        b = (
            exponent.unsqueeze(2).expand(-1, -1, self.bandwidths.shape[2])
            * -self.bandwidths
        )
        lossMMD = torch.sum(self.S.unsqueeze(2) * b.exp()).sqrt()

        return lossMMD




"""Implementations of edge clique clover finding algorithms."""
from .graph import UndirectedDependenceGraph
import numpy as np


def find_clique_min_cover(graph, verbose=False):
    """Returns the clique-minimum edge clique cover.

    Parameters
    ----------
    graph : 2d numpy array 
            Adjacency matrix for undirected graph.

    verbose : bool, optional
              Wether or not to print verbose output.

    Returns
    -------
    the_cover : 2d numpy array
                Biadjacency matrix representing edge clique cover.

    See Also
    --------
    graph.UndirectedDependenceGraph : Defines auxilliary data structure 
                                      and reduction rules used by this
                                      algorithm.

    Notes
    -----
    This is an implementation of the algorithm described in 
    :cite:`Gramm_2009`.

    """
    graph = UndirectedDependenceGraph(graph, verbose)
    try:
        graph.make_aux()
    except ValueError:
        print("The input graph doesn't appear to have any edges!")
        return graph.adj_matrix

    num_cliques = 1
    the_cover = None
    if True:# verbose:
        # find bound for cliques in solution
        max_intersect_num = graph.num_vertices ** 2 // 4
        if max_intersect_num < graph.num_edges:
            p = graph.n_choose_2(graph.num_vertices) - graph.num_edges
            t = int(np.sqrt(p))
            max_intersect_num = p + t if p > 0 else 1
        print("solution has at most {} cliques.".format(max_intersect_num))
    while the_cover is None:
        if True:                # verbose:
            print("\ntesting for solutions with {}/{} cliques".format(num_cliques, max_intersect_num))
        the_cover = branch(graph, num_cliques, the_cover)
        num_cliques += 1

    return the_cover


def branch(graph, k_num_cliques, the_cover):
    """Helper function for `find_clique_min_cover()`.

    Describing the solution search space as a tree, this function tests whether the given node is a solution, and it branches if not,

    Parameters
    ----------
    graph : UndirectedDependenceGraph()
            Class for representing undirected graph and auxilliary data used in edge clique cover algorithm.

    k_num_cliques : int
                    Current depth of search; number of cliques in cover being testet for solution.

    the_cover : 2d numpy array
                Biadjacency matrix representing (possibly partial) edge clique cover.

    Returns
    -------
    2d numpy array or None
        Biadjacency matrix representing (complete) edge clique cover or None if cover is only partial.

    """
    branch_graph = graph.reducible_copy()
    # if the_cover is not None:
    #     print(the_cover)
    #     for clique in the_cover:  # this might not be necessary, since the_cover_prime is only +1 clique
    #         print('clique: {}'.format(clique))
    #         branch_graph.the_cover = [clique]
    #         branch_graph.cover_edges()  # only works one clique at a time, or on a list of edges
    branch_graph.the_cover = the_cover
    branch_graph.cover_edges()
    
    if branch_graph.num_edges == 0:
        return branch_graph.reconstruct_cover(the_cover)

    # branch_graph.the_cover = the_cover

    branch_graph.reduzieren(k_num_cliques)
    k_num_cliques = branch_graph.k_num_cliques
    
    if k_num_cliques < 0:
        return None

    if branch_graph.num_edges == 0:  # equiv to len(branch_graph.extant_edges_idx)==0
        return branch_graph.the_cover  # not in paper, but speeds it up slightly; or rather return None?
    
    chosen_nbrhood = branch_graph.choose_nbrhood()
    # print("num cliques: {}".format(len([x for x in max_cliques(chosen_nbrhood)])))
    for clique_nodes in max_cliques(chosen_nbrhood):
        if len(clique_nodes) == 1:  # then this vert has been rmed; quirk of max_cliques
            continue
        clique = np.zeros(branch_graph.unreduced.num_vertices, dtype=int)
        clique[clique_nodes] = 1
        union = clique.reshape(1, -1) if branch_graph.the_cover is None else np.vstack((branch_graph.the_cover, clique))

        the_cover_prime = branch(branch_graph, k_num_cliques-1, union)
        if the_cover_prime is not None:
            return the_cover_prime
    return None


def max_cliques(nbrhood):
    """Adaptation of NetworkX code for finding all maximal cliques.

    Parameters
    ----------
    nbrhood : 2d numpy array
            Adjacency matrix for undirected (sub)graph.

    Returns
    -------
    generator
        set of all maximal cliques

    Notes
    -----
    Pieced together from nx.from_numpy_array and nx.find_cliques, which
    is output sensitive.

    """    
    if len(nbrhood) == 0:
        return
    
    # convert adjacency matrix to nx style graph
    adj = {u: {v for v in np.nonzero(nbrhood[u])[0] if v != u} for u in range(len(nbrhood))}
    Q = [None]

    subg = set(range(len(nbrhood)))
    cand = set(range(len(nbrhood)))
    u = max(subg, key=lambda u: len(cand & adj[u]))
    ext_u = cand - adj[u]
    stack = []

    try:
        while True:
            if ext_u:
                q = ext_u.pop()
                cand.remove(q)
                Q[-1] = q
                adj_q = adj[q]
                subg_q = subg & adj_q
                if not subg_q:
                    yield Q[:]
                else:
                    cand_q = cand & adj_q
                    if cand_q:
                        stack.append((subg, cand, ext_u))
                        Q.append(None)
                        subg = subg_q
                        cand = cand_q
                        u = max(subg, key=lambda u: len(cand & adj[u]))
                        ext_u = cand - adj[u]
            else:
                Q.pop()
                subg, cand, ext_u = stack.pop()
    except IndexError:
        pass    
    # note: max_cliques is a generator, so it's consumed after being
    # looped through once



""" Create and display various graphs, plots, and visualizations."""
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns

import networkx as nx



# %%
def show_obs_adj_mat(incidence_mat):

    fig, ax = plt.subplots()

    adj_mat = get_adj_from_incidence(incidence_mat)
    plt.imshow(adj_mat)

    dim = adj_mat.shape[0]

    ax.set_xticks(np.arange(0, dim))
    ax.set_xticklabels([str(idx) for idx in np.arange(1, dim + 1)])

    ax.set_yticks(np.arange(0, dim))
    ax.set_yticklabels([str(idx) for idx in np.arange(1, dim + 1)])

    plt.show()


# %%
def show_obs_dcor_mat(dcor_mat, thresh=None, print_val=False):

    fig, ax = plt.subplots()

    if thresh is not None:
        dcor_mat = (dcor_mat > thresh).astype(int)

    plt.imshow(dcor_mat, vmin=0, vmax=1)
    if print_val:
        for (j, i), label in np.ndenumerate(dcor_mat):
            ax.text(i, j, round(label, 2), ha='center', va='center', fontdict={'color': 'w'})

    dim = dcor_mat.shape[0]

    ax.set_xticks(np.arange(0, dim))
    ax.set_xticklabels([f'X{idx}' for idx in np.arange(1, dim + 1)])

    ax.set_yticks(np.arange(0, dim))
    ax.set_yticklabels([f'X{idx}' for idx in np.arange(1, dim + 1)])

    plt.colorbar()
    plt.show()


# %%
def show_dag(biadj_mat):
    """

    :param biadj_mat:
    :return:
    """

    num_latent, num_obs = biadj_mat.shape

    pos_dict = {}

    latent_pos_dict = {idx:(val,1) for idx, val in enumerate(np.linspace(0, 1, num_latent))}
    obs_pos_dict = {idx+num_latent:(val,0) for idx, val in enumerate(np.linspace(0, 1, num_obs))}


    pos_dict.update(latent_pos_dict)
    pos_dict.update(obs_pos_dict)
    # print(pos_dict)

    node_color =[]
    node_color.extend(num_latent*[0])
    node_color.extend(num_obs*[1])

    full_adj_mat = get_dag_from_biadj(biadj_mat)

    G = nx.DiGraph(full_adj_mat)

    nx.draw_networkx(G, pos=pos_dict, with_labels=False, node_size=2350)
    nx.draw_networkx_labels(G, pos=latent_pos_dict, labels={idx: '$L_{{{}}}$'.format(idx) for idx in range(num_latent)}, font_color='w')
    nx.draw_networkx_labels(G, pos=obs_pos_dict, labels={idx+num_latent: '$M_{{{}}}$'.format(idx) for idx in range(num_obs)}, font_color='k')
    nx.draw_networkx_nodes(G, node_size=2500, pos=pos_dict, node_color=node_color)
    # nx.draw_networkx(G, pos=pos_dict, arrows=True, with_labels=False)
    plt.xlim(-0.1, 1.1)
    plt.ylim(-0.5, 1.5)
    plt.show()

    
# %%
def get_dag_from_biadj(biadj_mat):

    num_latent, num_obs = biadj_mat.shape

    dag_adj_mat = np.zeros((num_latent+num_obs, num_latent+num_obs))

    dag_adj_mat[:num_latent, num_latent:] = biadj_mat

    return dag_adj_mat


# %%
def show_pairwise_plot(sample: np.ndarray, color='C0'):
    """

    :param sample:
    :return:
    """

    if sample.shape[0] > 1000:
        print(f'sample of size {sample.shape[0]} too big, using first 1000.')
        sample = sample[:1000, :]

    sample_df = pd.DataFrame(sample, columns=[f'X{idx}' for idx in np.arange(1, sample.shape[1]+1)])
    sns.pairplot(sample_df, corner=True, diag_kind='kde',
                 plot_kws=dict(color=color), diag_kws=dict(color=color))
    plt.show()



from setuptools import setup


def long_description():
    with open("README.md") as readme:
        return readme.read()


setup(
    name="medil",
    version="0.7.0",
    author="Alex Markham",
    author_email="alex.markham@causal.dev",
    description="This package is for causal modeling, originally focusing on the measurement dependence inducing latent (MeDIL) causal model framework, but now including more general methods for causal discovery and inference.",
    long_description_content_type="text/markdown",
    long_description=long_description(),
    license="Cooperative Non-Violent Public License v7 or later (CNPLv7+)",
    packages=["medil"],
    install_requires=["numpy"],
    extras_require={
        "dcor": ["dcor"],
        "GAN": ["pytorch-lightning"],
        "vis": ["matplotlib", "networkx"],
        "all": ["dcor", "torch", "matplotlib", "networkx"],
    },
)



"""Implement the Groebner basis-based UEC search algorithm (GrUES)."""
import numpy as np
import math
import warnings
from .gauss_obs_l0_pen import GaussObsL0Pen
from .independence_testing import dcov
from scipy.stats import chi2, beta
from scipy.special import factorial
from numpy.linalg import lstsq, det, inv


class InputData(object):
    r"""Feed data into GrUES to learn its UEC.

    Attributes
    ----------
    samples : array_like
        Passes directly to numpy.array to get a :math:`N \times M`
        matrix with :math:`N` samples of :math:`M` random variables.

    Methods
    -------
    grues()
        Return optimal undirected graph representing an unconditional
        equivalence class of DAGs.

    """

    def __init__(self, samples, rng=np.random.default_rng(0)):
        self.samples = np.array(samples, dtype=float)
        self.rng = rng
        self.pmf = None

        self.num_samps, self.num_feats = self.samples.shape
        self.explore = False
        self.debug = False

    def mcmc(self, init="empty", p="uniform", max_moves=10000, prior=None):
        if p == "uniform":
            self.p = {
                "merge": 1 / 6,
                "split": 1 / 6,
                "within": 1 / 3,
                "out_del": 1 / 6,
                "out_add": 1 / 6,
            }
        else:
            self.p = p
        self.init_uec(init)
        self.get_max_cpdag()
        self.reduce_max_cpdag()
        self.moves = 0
        self.visited = np.empty(max_moves, float)
        self.old_rss = self.compute_mle_rss()
        self.visited[0] = self.optimal_bic = self.cpdag.sum() * np.log(
            self.num_samps
        ) + self.num_samps * np.log(self.old_rss)
        self.markov_chain = np.empty((max_moves, self.num_feats, self.num_feats), bool)
        self.markov_chain[0] = self.uec
        while self.moves < max_moves - 1:
            self.old_bic = self.visited[self.moves]
            self.old_cpdag = np.copy(self.cpdag)
            self.old_dag = np.copy(self.dag_reduction)
            self.old_cc = np.copy(self.chain_comps)

            moves_dict = {
                "merge": self.merge,
                "split": self.split,
                "within": self.algebraic,
                "out_del": self.algebraic,
                "out_add": self.algebraic,
            }

            poss_moves, considered, p = self.compute_transition_kernel(moves_dict)
            move = self.rng.choice(poss_moves, p=p)

            q = float(self.q[move] * p[poss_moves == move])

            moves_dict[move](considered[move])

            if self.debug:
                try:
                    self.run_checks(move)
                except AssertionError:
                    import pdb, traceback

                    exc = traceback.format_exc()
                    pdb.set_trace()

            inv_moves = {
                "merge": "split",
                "split": "merge",
                "within": "within",
                "out_del": "out_add",
                "out_add": "out_del",
            }
            inv_move = inv_moves[move]

            poss_moves, _, p = self.compute_transition_kernel(moves_dict)
            q_inv = float(self.q[inv_move] * p[poss_moves == inv_move])

            likelihood_ratio, new_rss, new_bic = self.get_likelihood_ratio()
            if new_bic / self.optimal_bic < 1:
                self.optimal_bic = new_bic
                self.optimal_uec = self.uec

            likelihood_and_transition_ratio = likelihood_ratio * (q_inv / q)
            if type(prior) is tuple:
                prior_ratio = self.scaled_triangle_prior(prior)
                likelihood_and_transition_ratio *= prior_ratio
            elif prior is not None:
                prior_ratio = prior(self)
                likelihood_and_transition_ratio *= prior_ratio
            h = min(1, likelihood_and_transition_ratio)
            if self.explore:
                h = 1
            make_move = self.rng.choice((True, False), p=(h, 1 - h))
            if make_move:
                self.old_rss = new_rss
                self.moves += 1
                self.visited[self.moves] = new_bic
                self.markov_chain[self.moves] = self.uec
                # (2 ** np.flatnonzero(self.cpdag)).sum()]
            else:
                self.cpdag = self.old_cpdag
                self.dag_reduction = self.old_dag
                self.chain_comps = self.old_cc

    def init_uec(self, init):
        if type(init) is str:
            if init == "empty":
                self.uec = np.zeros((self.num_feats, self.num_feats), bool)
            elif init == "complete":
                self.uec = np.ones((self.num_feats, self.num_feats), bool)
                np.fill_diagonal(self.uec, False)
        elif type(init) is tuple:
            init, alpha = init
            if init == "gauss":
                corr = np.corrcoef(self.samples, rowvar=False)
                dist = beta(
                    self.num_samps / 2 - 1, self.num_samps / 2 - 1, loc=-1, scale=2
                )
                crit_val = abs(dist.ppf(alpha / 2))
                self.uec = self.indep_test_U = abs(corr) >= crit_val
                np.fill_diagonal(self.uec, False)
            elif init == "dcov_fast":
                cov, d_bars = dcov(self.samples)
                crit_val = chi2(1).ppf(1 - alpha)
                test_val = self.num_samps * cov / np.outer(d_bars, d_bars)
                self.uec = test_val >= crit_val
                np.fill_diagonal(self.uec, False)
        else:
            self.uec = np.array(init, bool)
        self.optimal_uec = self.uec

    def get_max_cpdag(self):
        r"""Return maximal CPDAG in the UEC."""

        U = np.copy(self.uec)
        compliment_U = ~U
        np.fill_diagonal(compliment_U, False)

        # V_ij == 1 if and only if there's a k adjacent to j but not i
        V = compliment_U @ U

        # W_ij == 1 if and only if there's k such that i--j--k is an induced path
        W = np.logical_and(V, U).T

        # This orients all v-structures and removes edges violating CI relations
        U[W] = False

        self.cpdag = U
        recon_uec = self.get_uec()
        if not (recon_uec == self.uec).all():
            # then self.uec was invalid, so reinitialize
            self.uec = self.optimal_uec = recon_uec
            self.get_max_cpdag()

    def reduce_max_cpdag(self):
        cpdag = np.copy(self.cpdag)
        undir = np.logical_and(cpdag, cpdag.T)
        chain_comps = np.eye(self.num_feats).astype(bool)
        while undir.any() and len(undir) > 1:
            v, w = np.unravel_index(undir.argmax(), undir.shape)
            cpdag = np.delete(cpdag, v, 0)
            cpdag = np.delete(cpdag, v, 1)
            undir = np.delete(undir, v, 0)
            undir = np.delete(undir, v, 1)
            chain_comps[w] += chain_comps[v]
            chain_comps = np.delete(chain_comps, v, 0)

        self.dag_reduction = cpdag
        self.chain_comps = chain_comps

    def merge(self, considered):
        src_1, src_2 = considered
        self.perform_merge(src_1, src_2)

    def perform_merge(self, src_1, src_2, recurse=True):
        self.chain_comps[src_1] += self.chain_comps[src_2]
        self.dag_reduction[src_1] += self.dag_reduction[src_2]

        self.chain_comps = np.delete(self.chain_comps, src_2, 0)
        self.dag_reduction = np.delete(self.dag_reduction, src_2, 0)
        self.dag_reduction = np.delete(self.dag_reduction, src_2, 1)

        if recurse:
            src_1 -= 1 if src_1 > src_2 else 0
            parentless = np.flatnonzero(self.dag_reduction.sum(0) == 1)
            for child in parentless:
                self.perform_merge(src_1, child, False)

    def split(self, considered):
        v, w, source = considered
        self.perform_split(v, w, source)

    def consider_split(self):
        # uniformly pick a source chain component to split
        source = self.pick_source_nodes("split")

        # uniformly pick edge v--w in the chain component to split on
        chain_comp_mask = self.chain_comps[source]
        choices = np.flatnonzero(chain_comp_mask)
        v, w = self.rng.choice(choices, 2, replace=False)

        self.q["split"] /= self.n_choose_2(len(choices))

        return v, w, source

    def perform_split(self, v, w, source, recurse=True, algebraic=False):
        # add node to dag reduction and corresponding cc to chain comps
        v_cc_mask = np.zeros((1, self.num_feats), bool)
        v_cc_mask[0, v] = True
        self.chain_comps[source, v] = False
        self.chain_comps = np.vstack((self.chain_comps, v_cc_mask))
        col = np.zeros((len(self.dag_reduction), 1), bool)
        self.dag_reduction = np.hstack((self.dag_reduction, col))
        # add edges from v_node to children of source node
        if not algebraic:
            self.dag_reduction = np.vstack(
                (self.dag_reduction, self.dag_reduction[source])
            )

        if recurse and self.chain_comps[source].sum() != 1:
            self.perform_split(w, None, source, False)
            self.dag_reduction[-2:, source] = True

    def algebraic(self, considered):
        src_1, src_2, t, v, T_mask = considered
        self.perform_algebraic(src_1, src_2, t, v, T_mask)

    def consider_algebraic(self, fiber):
        if fiber == "out_del":
            src_1, t = self.pick_source_nodes("out_del")
            src_2 = None
        else:
            src_1, src_2 = self.pick_source_nodes(fiber)
            ch_src_1, ch_src_2 = self.dag_reduction[[src_1, src_2], :]
            poss_t_mask = np.logical_and(ch_src_1, ~ch_src_2)
            poss_t_mask[src_1] = self.chain_comps[src_1].sum() > 1
            choices = np.flatnonzero(poss_t_mask)
            t = self.rng.choice(choices)
            self.q[fiber] /= len(choices)
            if len(choices) == 0:
                warnings.warn("inaccurate self.q")
        choices = np.flatnonzero(self.chain_comps[t])
        v = self.rng.choice(choices)
        self.q[fiber] /= len(choices)
        if len(choices) == 0:
            warnings.warn("inaccurate self.q")

        if src_1 == t:
            T_mask = np.zeros(len(self.dag_reduction), bool)
            T_mask[t] = True
        else:
            max_anc_mask = self.dag_reduction.sum(0) == 0
            par_t_mask = self.dag_reduction[:, t]
            T_mask = np.logical_and(max_anc_mask, par_t_mask)

        if fiber == "out_del":
            T_mask[src_1] = False
        elif fiber == "out_add":
            T_mask[src_2] = True
        else:  # fiber == "within"
            T_mask[[src_1, src_2]] = False, True

        return src_1, src_2, t, v, T_mask

    def perform_algebraic(self, src_1, src_2, t, v, T_mask):
        nonsources_mask = self.dag_reduction.sum(0).astype(bool)
        max_anc_dag = np.copy(self.dag_reduction)
        max_anc_dag[nonsources_mask] = False
        sources = np.flatnonzero(np.logical_not(nonsources_mask))
        max_anc_dag[sources, sources] = True

        num_common = T_mask.astype(int) @ max_anc_dag
        has_other_ancs = ~T_mask @ max_anc_dag
        P_mask = np.logical_and(num_common.astype(bool), ~has_other_ancs)
        C_mask = num_common == T_mask.sum()
        exact = np.flatnonzero(np.logical_and(P_mask, C_mask))
        if len(exact) == 1:
            self.chain_comps[[exact[0], t], v] = True, False
        else:
            self.perform_split(v, None, t, recurse=False, algebraic=True)
            C_mask = np.append(C_mask, False)
            self.dag_reduction = np.vstack((self.dag_reduction, C_mask))
            P_mask = np.append(P_mask, False)
            self.dag_reduction[P_mask, -1] = True

        if not self.chain_comps[t].any():
            self.chain_comps = np.delete(self.chain_comps, t, 0)
            self.dag_reduction = np.delete(self.dag_reduction, t, 0)
            self.dag_reduction = np.delete(self.dag_reduction, t, 1)

    def pick_source_nodes(self, move):
        # sources have no parents; sinks have parents and no children
        non_srcs_mask = self.dag_reduction.sum(0).astype(bool)
        sources = np.flatnonzero(np.logical_not(non_srcs_mask))
        childless_mask = np.logical_not(self.dag_reduction.sum(1).astype(bool))
        sinks = np.flatnonzero(np.logical_and(non_srcs_mask, childless_mask))
        if move == "merge":
            singleton_nodes = np.flatnonzero(self.chain_comps.sum(1) == 1)
            sngl_srcs = sources[np.in1d(sources, singleton_nodes)]
            ch_subgraph = self.dag_reduction[sngl_srcs].astype(int)
            empty_ch_subgraph = (~self.dag_reduction[sngl_srcs]).astype(int)
            non_empty_same_ch_mask = ch_subgraph @ ch_subgraph.T
            empty_same_ch_mask = empty_ch_subgraph @ empty_ch_subgraph.T
            same_ch_mask = non_empty_same_ch_mask + empty_same_ch_mask
            np.fill_diagonal(same_ch_mask, 0)
            same_ch_idx = np.argwhere(same_ch_mask == len(self.dag_reduction))
            choices = range(len(same_ch_idx))
            idx = self.rng.choice(choices)
            self.q[move] /= len(choices)
            src_1, src_2 = sngl_srcs[same_ch_idx[idx]]
            chosen_nodes = src_1, src_2
        elif move == "out_del":
            num_max_ancs = self.dag_reduction[sources].sum(0)
            num_pairs = self.n_choose_2(num_max_ancs)
            poss_t = np.flatnonzero(num_pairs)
            p = num_pairs[poss_t] / num_pairs[poss_t].sum()
            t = self.rng.choice(poss_t, p=p)
            self.q[move] *= p[poss_t == t]
            t_max_ancs = np.flatnonzero(self.dag_reduction[sources, t])
            src_1 = sources[self.rng.choice(t_max_ancs)]
            self.q[move] /= len(t_max_ancs)
            if len(t_max_ancs) == 0:
                warnings.warn("inaccurate self.q")
            chosen_nodes = src_1, t
        else:  # then move in ("split", "within", "out_add")
            non_singleton_nodes = np.flatnonzero(self.chain_comps.sum(1) > 1)
            ns_sources = sources[np.in1d(sources, non_singleton_nodes)]
            if move == "split":
                chosen_nodes = self.rng.choice(ns_sources)
                self.q[move] /= len(ns_sources)
            else:  # move in ("within", "out_add")
                # srcs_mask has entry i,j = 1 if and only if
                # i is nonsingleton or
                # i has children other than j's children
                other_children_mask = self.dag_reduction @ ~self.dag_reduction.T
                other_children_mask[ns_sources] = 1
                srcs_mask = other_children_mask[np.ix_(sources, sources)]
                np.fill_diagonal(srcs_mask, 0)
                choices = range(srcs_mask.sum())
                chosen_idx = self.rng.choice(choices)
                self.q[move] /= len(choices)
                chosen_nodes = sources[np.argwhere(srcs_mask)[chosen_idx]]
        return chosen_nodes

    @staticmethod
    def n_choose_2(vec):
        return np.vectorize(lambda n: math.comb(n, 2), otypes=[int])(
            np.array(vec, ndmin=1)
        )

    def expand(self):
        self.cpdag = np.zeros_like(self.cpdag)
        for pa in np.flatnonzero(self.dag_reduction.sum(1)):
            pas = self.chain_comps[pa]
            ch = self.dag_reduction[pa]
            chs = self.chain_comps[ch].sum(0).astype(bool)
            self.cpdag[np.ix_(pas, chs)] = True
        for cc in self.chain_comps:
            nodes = np.flatnonzero(cc)
            for node in nodes:
                ch = nodes[nodes != node]
                self.cpdag[node, ch] = True
        np.fill_diagonal(self.cpdag, False)

    def get_uec(self, cpdag=None):
        if cpdag is None:
            G = self.cpdag
        else:
            G = cpdag
        T = np.eye(len(G)).astype(bool) + G + np.linalg.matrix_power(G, 2)
        recon_uec = T.T @ T
        np.fill_diagonal(recon_uec, False)
        return recon_uec

    def get_likelihood_ratio(self):
        self.expand()
        new_rss = self.compute_mle_rss()
        new_bic = self.num_samps * np.log(new_rss)
        new_bic += self.cpdag.sum() * np.log(self.num_samps)
        # unscaled_ratio = (new_likelihood / self.old_likelihood).astype(np.longdouble)
        return self.old_rss / new_rss, new_rss, new_bic

    def compute_mle_rss(self, graph=None):
        if graph is None:
            graph = self.cpdag
            self.uec = self.get_uec()
        # children_mask = graph.sum(0)
        # children = np.flatnonzero(children_mask)

        # regress = lambda child: lstsq(
        #     self.samples[:, graph[:, child]],
        #     self.samples[:, child],
        #     rcond=None,
        # )[1]
        # rss = float(sum(map(regress, children)))
        # non_children = np.flatnonzero(~children_mask)
        # nc_samps = self.samples[:, non_children]
        # rss += np.diag(nc_samps.T @ nc_samps).sum()
        # # rss here is just n * var(x), which is equiv to the above since data is centered

        # return rss
        return GaussObsL0Pen(self.samples)._mle_full(graph)[1].sum()

    def run_checks(self, move):
        # dag and ccs are correct type
        assert self.chain_comps.dtype == bool
        assert self.dag_reduction.dtype == bool

        # each vertex is a chain component
        assert len(self.chain_comps) == len(self.dag_reduction)
        assert (self.chain_comps.sum(0) == 1).all()

        # every child has a unique parent set with at least two parents
        num_pars = self.dag_reduction.sum(0)
        children = self.dag_reduction[:, num_pars.astype(bool)]
        assert np.unique(children, axis=1).shape[1] == children.shape[1]
        assert (num_pars != 1).all()

        # indeed a DAG and trasitively closed
        num_nodes = len(self.dag_reduction)
        graph = np.copy(self.dag_reduction).astype(int)
        for n in range(2, num_nodes):
            graph += np.linalg.matrix_power(self.dag_reduction, n)
        assert np.diag(graph).sum() == 0
        assert (graph.astype(bool) == self.dag_reduction).all()

        # all edges are essential
        # i.e., all edges are in v-structs
        A = self.dag_reduction
        comp_undir = ~(A + A.T)
        np.fill_diagonal(comp_undir, False)
        assert np.logical_or(~A, comp_undir @ A).all

        # check interesection number
        old_intersection_num = np.logical_not(self.old_dag.sum(0)).sum()
        new_intresection_num = np.logical_not(self.dag_reduction.sum(0)).sum()
        if move == "merge":
            assert old_intersection_num - 1 == new_intresection_num
        elif move == "split":
            assert old_intersection_num + 1 == new_intresection_num
        elif move == "algebraic":
            assert old_intersection_num == new_intresection_num

    def sort(self):
        dag = np.copy(self.dag_reduction) + np.eye(len(self.dag_reduction))
        sorted_idx = np.array([], int)
        while dag.any():
            sink_ccs = np.flatnonzero(dag.sum(1) == 1)
            sinks = np.where(self.chain_comps[sink_ccs])[1]
            sorted_idx = np.append(sinks, sorted_idx)
            dag[sinks] = dag[:, sinks] = 0
        return sorted_idx

    def consider(self, move):
        if move == "merge":
            return self.pick_source_nodes("merge")
        elif move == "split":
            return self.consider_split()
        elif move == "within":
            return self.consider_algebraic("within")
        elif move == "out_del":
            return self.consider_algebraic("out_del")
        elif move == "out_add":
            return self.consider_algebraic("out_add")

    def compute_transition_kernel(self, moves_dict):
        poss_moves = []
        p = []
        considered = {}
        self.q = {key: 1.0 for key in self.p.keys()}
        for move in moves_dict.keys():
            try:
                considered[move] = self.consider(move)
                poss_moves += [move]
                p += [self.p[move]]
            except ValueError:
                continue
        p = np.array(p)
        p /= p.sum()

        return np.array(poss_moves), considered, p

    def scaled_triangle_prior(self, peak_scale):
        if self.pmf is None:
            peak, scale = peak_scale
            min = 0
            max = self.num_feats + 1
            pmf = np.arange(1, max, dtype=float)
            pmf[peak - 1] = 2 / max
            below = pmf[: peak - 1]
            above = pmf[peak:]

            below *= 2 / (max * peak)
            above *= -1
            above += max
            above *= 2 / (max * (max - peak))
            pmf[: peak - 1] = below
            pmf[peak:] = above

            pmf = np.power(pmf, scale)
            pmf /= pmf.sum()
            self.pmf = pmf

        new_num_sources = (self.dag_reduction.sum(0) == 0).sum()
        old_num_sources = (self.old_dag.sum(0) == 0).sum()

        return self.pmf[new_num_sources - 1] / self.pmf[old_num_sources - 1]



# Copyright 2021 Juan L Gamella

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
"""

import numpy as np
from .decomposable_score import DecomposableScore

# --------------------------------------------------------------------
# l0-penalized Gaussian log-likelihood score for a sample from a single
# (observational) environment


class GaussObsL0Pen(DecomposableScore):
    """
    Implements a cached l0-penalized gaussian likelihood score.
    """

    def __init__(self, data, lmbda=None, method="scatter", cache=True, debug=0):
        """Creates a new instance of the class.
        Parameters
        ----------
        data : numpy.ndarray
            the nxp matrix containing the observations of each
            variable (each column corresponds to a variable).
        lmbda : float or NoneType, optional
            the regularization parameter. If None, defaults to the BIC
            score, i.e. lmbda = 1/2 * log(n), where n is the number of
            observations.
        method : {'scatter', 'raw'}, optional
            the method used to compute the likelihood. If 'scatter',
            the empirical covariance matrix (i.e. scatter matrix) is
            used. If 'raw', the likelihood is computed from the raw
            data. In both cases an intercept is fitted.
        cache : bool, optional
           if computations of the local score should be cached for
           future calls. Defaults to True.
        debug : int, optional
            if larger than 0, debug are traces printed. Higher values
            correspond to increased verbosity.
        """
        if type(data) != np.ndarray:
            raise TypeError("data should be numpy.ndarray, not %s." % type(data))

        super().__init__(data, cache=cache, debug=debug)

        self.n, self.p = data.shape
        self.lmbda = 0.5 * np.log(self.n) if lmbda is None else lmbda
        self.method = method

        # Precompute scatter matrices if necessary
        if method == "scatter":
            self._scatter = np.cov(data, rowvar=False, ddof=0)
        elif method == "raw":
            self._centered = data - np.mean(data, axis=0)
        else:
            raise ValueError('Unrecognized method "%s"' % method)

    def full(self, A):
        """
        Given a DAG adjacency A, return the l0-penalized log-likelihood of
        a sample from a single environment, by finding the maximum
        likelihood estimates of the corresponding connectivity matrix
        (weights) and noise term variances.
        Parameters
        ----------
        A : np.array
            The adjacency matrix of a DAG, where A[i,j] != 0 => i -> j.
        Returns
        -------
        score : float
            the penalized log-likelihood score.
        """
        # Compute MLE
        B, omegas = self._mle_full(A)
        # Compute log-likelihood (without log(2π) term)
        K = np.diag(1 / omegas)
        I_B = np.eye(self.p) - B.T
        log_term = self.n * np.log(omegas.prod())
        if self.method == "scatter":
            # likelihood = 0.5 * self.n * (np.log(det_K) - np.trace(K @ I_B @ self._scatter @ I_B.T))
            likelihood = log_term + self.n * np.trace(K @ I_B @ self._scatter @ I_B.T)
        else:
            # Center the data, exclude the intercept column
            inv_cov = I_B.T @ K @ I_B
            cov_term = 0
            for i, x in enumerate(self._centered):
                cov_term += x @ inv_cov @ x
            likelihood = log_term + cov_term
        #   Note: the number of parameters is the number of edges + the p marginal variances
        l0_term = self.lmbda * (np.sum(A != 0) + 1 * self.p)
        score = -0.5 * likelihood - l0_term
        return score

    # Note: self.local_score(...), with cache logic, already defined
    # in parent class DecomposableScore.

    def _compute_local_score(self, x, pa):
        """
        Given a node and its parents, return the local l0-penalized
        log-likelihood of a sample from a single environment, by finding
        the maximum likelihood estimates of the weights and noise term
        variances.
        Parameters
        ----------
        x : int
            a node.
        pa : set of ints
            the node's parents.
        Returns
        -------
        score : float
            the penalized log-likelihood score.
        """
        pa = list(pa)
        # Compute MLE
        b, sigma = self._mle_local(x, pa)
        # Compute log-likelihood (without log(2π) term)
        likelihood = -0.5 * self.n * (1 + np.log(sigma))
        #  Note: the number of parameters is the number of parents (one
        #  weight for each) + the marginal variance of x
        l0_term = self.lmbda * (len(pa) + 1)
        score = likelihood - l0_term
        return score

    # --------------------------------------------------------------------
    #  Functions for the maximum likelihood estimation of the
    #  weights/variances

    def _mle_full(self, A):
        """
        Finds the maximum likelihood estimate for the whole graph,
        specified by the adjacency A.
        Parameters
        ----------
        A : np.array
            The adjacency matrix of a DAG, where A[i,j] != 0 => i -> j.
        Returns
        -------
        B : np.array
            the connectivity (weights) matrix, which respects the
            adjacency in A.
        omegas : np.array
            the estimated noise-term variances of the observed
            variables.
        """
        B = np.zeros(A.shape)
        omegas = np.zeros(self.p)
        for j in range(self.p):
            parents = np.where(A[:, j] != 0)[0]
            B[:, j], omegas[j] = self._mle_local(j, parents)
        return B, omegas

    def _mle_local(self, j, parents):
        """Finds the maximum likelihood estimate of the local model
        between a node and its parents.
        Parameters
        ----------
        x : int
            a node.
        pa : set of ints
            the node's parents.
        Returns
        -------
        b : np.array
            an array of size p, with the estimated weights from the
            parents to the node, and zeros for non-parent variables.
        sigma : float
            the estimate noise-term variance of variable x
        """
        parents = list(parents)
        b = np.zeros(self.p)
        # Compute the regression coefficients from a least squares
        # regression on the raw data
        if self.method == "raw":
            Y = self._centered[:, j]
            if len(parents) > 0:
                X = np.atleast_2d(self._centered[:, parents])
                # Perform regression
                coef = np.linalg.lstsq(X, Y, rcond=None)[0]
                b[parents] = coef
                sigma = np.var(Y - X @ coef)
            else:
                sigma = np.var(Y, ddof=0)
        # Or compute the regression coefficients from the
        # empirical covariance (scatter) matrix
        # i.e. b = Σ_{j,pa(j)} @ Σ_{pa(j), pa(j)}^-1
        elif self.method == "scatter":
            sigma = self._scatter[j, j]
            if len(parents) > 0:
                cov_parents = self._scatter[parents, :][:, parents]
                cov_j = self._scatter[j, parents]
                # Use solve instead of inverting the matrix
                coef = np.linalg.solve(cov_parents, cov_j)
                sigma = sigma - cov_j @ coef
                b[parents] = coef
        # For very low sample-sizes sigma can become lower than or
        # zero -> set to machine epsilon to avoid log(0) in the
        # likelihood computation
        if sigma <= 0:
            sigma = abs(np.finfo(float).eps)
        return b, sigma



"""Independence testing on samples of random variables."""
import numpy as np
from scipy.spatial.distance import pdist, squareform, cdist
from multiprocessing import Pool

from numpy import linalg as LA
from scipy.stats import chi2

try:
    from dcor import pairwise, distance_correlation as dist_corr

    default_measure = "dcor"
except ImportError:
    default_measure = "pearson"


def hypothesis_test(samples, num_resamples, measure=default_measure, alpha=0.05):
    r"""Performs random permutation tests to estimate independence and
    returns the estimated Undirected Dependency Graph in the form of an
    adjacency matrix.

    Parameters
    ----------
    samples : 2d numpy array of floats or ints
              A :math:`M \times N` matrix with :math:`N` samples of
              :math:`M` random variables.

    num_resamples : int, optional
                    Number of permutations used to calculate p-values.

    measure : str, optional
              Measure of association to use as test statistic. Either
              `pearson` (default if `dcor` package is not installed)
              for a linear measure or `dcor` (default if installed) for
              a nonlinear measure.

    alpha : float, optional
            The threshold on the p-values above which a result is
            considered statistically significant.

    Returns
    -------
    p_values : 2d numpy array of floats
               A square matrix :math:`P`, where :math:`P_{i,j}` is the
               probability of obtaining a result at least as extreme
               as the one given by :math:`N_{i, j}`.

    sample_corr : 2d numpy array of floats
                  A square matrix :math:`N`, where :math:`N_{i,j}` is
                  the measured association value (e.g., correlation)
                  between random variables :math:`R_i` and :math:`R_j`.

    deps : 2d numpy array of bools
           A square matrix `D`, where :math:`D_{i,j}` is true if and
           only if the corresponding random variables :math:`R_i` and
           :math:`R_j` are not estimated to be independent.

    See Also
    --------
    pearson_correlation : Used when ``measure == 'pearson'``.
    distance_correlation : Used when ``measure == 'dcor'``.

    Notes
    -----
    The runtime when using ``distance_correlation`` is nearly cut in
    halve by first generating two permuted samples matrices, ``a`` and
    ``b`` and then calling ``distance_correlation(a, b)`` and using the
    upper and lower triangles, as opposed to calling
    ``distance_correlation(a)`` and ``distance_correlation(b)``
    separately while using only one triangle from each. A similar trick
    could be employed in ``pearson_correlation``, but it reduces the
    runtime by less than 10%. Another (mathematically) similar trick
    would be to compute the correlation on ``np.vstack((a, b))``, but
    this also is less than a 10% improvement.

    """
    if measure == "pearson":
        compute_corr = pearson_correlation
    elif measure == "dcor":
        compute_corr = distance_correlation
    else:
        raise ValueError("{} is not a supported measure of association".format(measure))

    sample_corr = compute_corr(samples)

    # initialize aux vars used in loop
    p_values = np.zeros(sample_corr.shape)
    num_loops = num_resamples if measure != "dcor" else int(np.ceil(num_resamples / 2))
    for _ in range(num_loops):
        perm_corr = compute_corr(samples, perm=True)
        p_values += np.array(perm_corr >= sample_corr, int)

    p_values += p_values.T
    p_values /= num_loops if measure != "dcor" else 2 * num_loops

    deps = p_values <= alpha  # reject null hypothesis of independence

    return p_values, sample_corr, deps


def distance_correlation(samples, perm=False):
    r"""Compute distance correlation on samples set (or on permuted samples
    set, if ``perm`` is ``True).

    Parameters
    ----------
    samples : 2d numpy array of floats or ints
              A :math:`M \times N` matrix with :math:`N` samples of
              :math:`M` random variables.

    perm : bool, optional
           Whether distance correlation is computed on permuted or
           original samples.

    Returns
    -------
    2d numpy array of floats
        A square matrix :math:`C`, where :math:`C_{i,j}` is
        (if ``perm``, a sample from the null distribution of)
        the distance correlation between random variables :math:`R_i`
        and :math:`R_j`.

    """
    try:
        pairwise
    except NameError:
        raise ImportError(
            "Install the dcor package to use this feature: `pip install dcor`."
        )
    if not perm:
        with Pool() as pool:
            corr = pairwise(dist_corr, samples, pool=pool)
    else:
        samples_permed = permute_within_rows(samples)
        samples_permed_2 = permute_within_rows(samples)
        with Pool() as pool:
            corr = pairwise(dist_corr, samples_permed, samples_permed_2, pool=pool)
    return corr


def pearson_correlation(samples, perm=False):
    r"""Computes Pearson product-moment correlation coefficient on samples
    set (or on permuted samples set, if ``perm`` is ``True).

    Parameters
    ----------
    samples : 2d numpy array of floats or ints
              A :math:`M \times N` matrix with :math:`N` samples of
              :math:`M` random variables.

    perm : bool, optional
           Whether Pearson correlation is computed on permuted or
           original samples.

    Returns
    -------
    2d numpy array of floats
        A square matrix :math:`C`, where :math:`C_{i,j}` is
        (if ``perm``, a sample from the null distribution of)
        the Pearson correlation between random variables :math:`R_i`
        and :math:`R_j`.

    """
    if not perm:
        corr = np.corrcoef(samples)
    else:
        samples_permed = permute_within_rows(samples)
        corr = np.corrcoef(samples, samples_permed)
    return corr


def permute_within_rows(x):
    """Randomly rearrange values according to column index without
    changing row index.

    Parameters
    ----------
    x : 2d numpy array

    Returns
    -------
    2d numpy array
        Reordered copy of input.

    """
    # get random new index for col of each element
    col_idx = np.random.sample(x.shape).argsort(axis=1)

    # keep the row index the same
    row_idx = np.tile(np.arange(x.shape[0]), (x.shape[1], 1)).T

    # apply the permutaton matrix to permute x
    return x[row_idx, col_idx]


def dcov(samples):
    r"""Compute sample distance covariance matrix.

    Parameters
    ----------
    samples : 2d numpy array of floats
              A :math:`N \times M` matrix with :math:`N` samples of
              :math:`M` random variables.

    Returns
    -------
    2d numpy array
        A square matrix :math:`C`, where :math:`C_{i,j}` is the sample
        distance covariance between random variables :math:`R_i` and
        :math:`R_j`.

    """
    num_samps, num_feats = samples.shape
    num_pairs = num_samps * (num_samps - 1) // 2
    dists = np.zeros((num_feats, num_pairs))
    # compute doubly centered distance matrix for every feature:
    for feat_idx in range(num_feats):
        n = num_samps
        t = np.tile
        # raw distance matrix:
        d = squareform(pdist(samples[:, feat_idx].reshape(-1, 1), "cityblock"))
        # doubly centered:
        d -= t(d.mean(0), (n, 1)) + t(d.mean(1), (n, 1)).T - t(d.mean(), (n, n))
        d = squareform(d, checks=False)  # ignore assymmetry due to numerical error
        dists[feat_idx] = d
    return dists @ dists.T / num_samps**2
